/*********************************************************************
 * 
 *    Filename:     cpa.c
 *    Author:       Ryan Shillington <shillington@writeme.com>
 *    Description:  Corel Palm Administrator
 *    Details:      This application is for Palm Pilot.  It acts as a
 *			client to get/set informtion on the Netwinder.
 *  
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of
 *    the License, or (at your option) any later version.
 *
 ********************************************************************/

#pragma pack(2)

register void *reg_a4 asm("%a4");

#define CALLBACK_PROLOGUE \
    void *save_a4 = reg_a4; asm("move.l %%a5,%%a4; sub.l #edata,%%a4" : :);

#define CALLBACK_EPILOGUE reg_a4 = save_a4;

#include "cpa.h"
#include "telnet.h"
#include <Common.h>
#include <irlib.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include <System/Unix/sys_socket.h>

//* Variables (for Execute) to print to the screen in line mode
#define TEXT_TOP   15
#define TEXT_LINES 9
#define MAX_LEN   80
#define MAX_LINES 80
char text_buffer[MAX_LINES+1][MAX_LEN+1];
int  text_height;
WinHandle draw_win;
int disp_line;

//* Used to scroll the EXECUTE screen up and down
void EAScrollDown(void);
void EAScrollUp(void);

//* The number of the current form
int currentForm;

//* Functions (for Execute) to print to the screen in line mode
void init_buffer(void);
void debug_msg(char *msg);
#define PRINTF(format, args...) { char tmp[MAX_LEN]; StrPrintF(tmp, format, ## args); debug_msg(tmp); }

//* These function are called to handle certain buttons being pressed
//* The first character describes where the button is (ie. 'S'-System Info,
//* 'M' Main Menu, etc.);
void SATemperature();
void SAMemory();
void SALoad();
void MChangeIP();
void CChangeIP(char port);
void MExecute();
//* Execute the command in the field
void EAExecute();
//* Handles a key being pressed in the telnet deamon
void Tprint(char input);
//* Handles incomming IR requests/confirmations
void IrCallback(IrConnect *connect, IrCallBackParms *parms);
//* Open the IR port
void openIR(void);
//* Handles GUI events
void eventLoop(void);
//* Used to send data to the server
void sendData(char *data, int Len);
void sendData2(char *data, int Len);
//* Used to handle incomming data
void handleData(char *data, int len);
//* Puts commas every 3 characters when printing
void PrintNUMwCommas(char *data, int Len, int x, int y);
//* This sets the text in a field (it's not as easy as it looks)
void SetFieldTextFromStr(Word fieldID, CharPtr strP);
//* Set the IP information on the netwinder
void CASetIP();
//* INDIVIDUAL HANDLERS (FOR INCOMMING DATA)
void STemperatureHandler(char *data, int len);
void SMemoryHandler(char *data, int len);
void SLoadHandler(char *data, int len);
void ChangeIPHandler(char *data, int len);
void ExecuteHandle(char *data, int len);

#define CPACreator	0xCCC 		// Creator ID for CPA

//* Library reference number
UInt irLibRef=0;
//* IR Connection information
IrConnect con;
//* Netwinder(Server) IR address
IrDeviceAddr dev_addr;
//* Packet is a wrapper for data being sent
IrPacket packet;
//* Information for Palm's database
IrIasObject ias_object;
char dev_name[]={IR_HINT_PDA, 0, 'P','a','l','m','E'};
char ias_name[]="CPA";
//* A pointer to an array of attributes
IrIasAttribute attribs;
char attr_name[]="IrDA:TinyTP:LsapSel";
char attr_value[]={ IAS_ATTRIB_INTEGER, 0,0,0,42};

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
  if (cmd == sysAppLaunchCmdNormalLaunch)
	{
		openIR();
		eventLoop();
	}
	return 0;
}

void eventLoop(void)
{
	Err err;
	EventType e;
	FormType *pfrm;
	unsigned long len;
	
	currentForm = SplashScreen;
	FrmGotoForm(currentForm);
	do {
		EvtGetEvent(&e, 100);
		if (SysHandleEvent(&e)) 
			continue;
		if (MenuHandleEvent((void *)0, &e, &err)) 
			continue;

		switch (e.eType) {
		case ctlSelectEvent:
      switch (e.data.ctlSelect.controlID) {
				case Main:
					FrmGotoForm(Main);
					break;
				case SystemInfo:
					FrmGotoForm(SystemInfo);
					break;
				case ChangeIP:
					FrmGotoForm(ChangeIP);
					break;
				case Execute:
					FrmGotoForm(Execute);
					break;
				case STemperature:
					FrmGotoForm(e.data.ctlSelect.controlID);
					break;
				case SMemory:
					FrmGotoForm(e.data.ctlSelect.controlID);
					break;
				case SLoadAvg:
					FrmGotoForm(e.data.ctlSelect.controlID);
					break;
				case EExecute:
					EAExecute();
					break;
				case CSetIP:
					CASetIP();
					break;
				case Die:
					return;
				case PDEVICE0:
					CChangeIP('0');
					break;
				case PDEVICE1:
					CChangeIP('1');
					break;
			}
			currentForm = e.data.ctlSelect.controlID;
			goto Dft;
			break;
		case frmLoadEvent:
	        	FrmSetActiveForm(FrmInitForm(e.data.frmLoad.formID));
			break;

		case frmOpenEvent:
			pfrm = FrmGetActiveForm();
			FrmDrawForm(pfrm);
			switch (e.data.frmOpen.formID) {
				case STemperature: 
					SATemperature();
					break;
				case SMemory: 
					SAMemory();
					break;
				case ChangeIP: 
					MChangeIP();
					break;
				case Execute:
					MExecute();
					break;
				case SLoadAvg: 
					SALoad();
					break;
			}
			break;
		
		case menuEvent:
			FrmAlert(kidAlert1);
			if(irLibRef != 0) IrClose(irLibRef);
			return;
			break;

		case appStopEvent:
			if(irLibRef != 0) IrClose(irLibRef);
			return;

  	case ctlRepeatEvent:
    	if(e.data.ctlEnter.controlID == EScrollUp)
				EAScrollUp();

    	if(e.data.ctlEnter.controlID == EScrollDown)
					EAScrollDown();
    break;

		default:
Dft:
			FrmHandleEvent(FrmGetActiveForm(), &e);
		}
	} while (e.eType != appStopEvent);
}

void MExecute()
{
	//* Set the window to draw in
  draw_win=WinGetDrawWindow();
	
	//* Initialize the Buffer information
	init_buffer();
}

void EAExecute()
{
	char data[256];
	int device;
	FormPtr thisForm = FrmGetActiveForm();
	
	if(FldGetTextPtr(FrmGetObjectPtr(thisForm, 
					FrmGetObjectIndex(thisForm, FCommand))) != NULL)
	{
		//* Properly format the data to send (pull out the data from the fields)
		StrPrintF(data, "Execute=%s\n", 
				FldGetTextPtr(FrmGetObjectPtr(thisForm, 
						FrmGetObjectIndex(thisForm, FCommand))));
	
		//* Send the data
		sendData(data, StrLen(data));
	}
}

void CASetIP()
{
	char data[256];
	int device;
	FormPtr thisForm = FrmGetActiveForm();
	
	//* Decide which device is currently selected
	if (CtlGetValue(FrmGetObjectPtr(thisForm, FrmGetObjectIndex(thisForm, PDEVICE0))))
		device = PDEVICE0;
	else 
		device = PDEVICE1;


	//* Properly format the data to send (pull out the data from the fields)
	StrPrintF(data, "SChangeIP=%s,%s,%s,%s,%s,%s,\n", 
			CtlGetLabel(FrmGetObjectPtr(thisForm, 
					FrmGetObjectIndex(thisForm, device))),
			FldGetTextPtr(FrmGetObjectPtr(thisForm, 
					FrmGetObjectIndex(thisForm, FIPADDRESS))),
			FldGetTextPtr(FrmGetObjectPtr(thisForm, 
					FrmGetObjectIndex(thisForm, FNAME))),
			FldGetTextPtr(FrmGetObjectPtr(thisForm, 
					FrmGetObjectIndex(thisForm, FALIAS))),
			FldGetTextPtr(FrmGetObjectPtr(thisForm, 
					FrmGetObjectIndex(thisForm, FNETMASK))),
			FldGetTextPtr(FrmGetObjectPtr(thisForm, 
					FrmGetObjectIndex(thisForm, FGATEWAY))));

	//* Send the data
	sendData2(data, StrLen(data));
}

void MChangeIP()
{
	int find, Fcurr = 0;
	int Ccurr = 0;

	//* 2 - Send our request to the server
	CChangeIP('?');
}

void CChangeIP(char port)
{
	char data[32];

	StrPrintF(data, "ChangeIP-eth%c.\n", port);
	sendData(data, StrLen(data));
}
	
void SATemperature()
{
	int len;

	//* 1 - Display what we know for sure
	WinDrawChars("Max Temperature:", 16, 10, 50);
	WinDrawChars("Min Temperature:", 16, 10, 60);
	WinDrawChars("Current Temperature:", 20, 10, 70);
	WinDrawChars("Fan:", 4, 10, 80);
	
	//* 2 - Send our request to the server
	sendData("Temperature.\n", 13);
}

void SALoad()
{
	int len;

	//* 1 - Display what we know for sure
	WinDrawChars("System Time:", 12, 10, 50);
	WinDrawChars("System has been up:", 19, 10, 60);
	WinDrawChars("Number of users:", 16, 10, 70);
	WinDrawChars("1 min Load Average:", 19, 10, 80);
	WinDrawChars("5 min Load Average:", 19, 10, 90);
	WinDrawChars("15 min Load Average:", 20, 10, 100);
	
	//* 2 - Send our request to the server
	sendData("Load.\n", 10);
}

void SAMemory()
{
	int len;

	//* 1 - Display what we know for sure
	//* 1.1 - Display row titles
	WinDrawChars("Total:", 6, 10, 60);
	WinDrawChars("Used:", 5, 10, 70);
	WinDrawChars("Free:", 5, 10, 80);
	WinDrawChars("Shared:", 7, 10, 90);
	WinDrawChars("Buffers:", 8, 10, 100);
	WinDrawChars("Cached:", 7, 10, 110);
	//* 1.2 - Display column titles
	WinDrawChars("Physical:", 9, 50, 50);
	WinDrawChars("Swap:", 5, 100, 50);
	
	//* 2 - Send our request to the server
	sendData("Memory.\n", 10);
}

void sendData(char *data, int Len)
{
	//* 1 - Put a message on the screen so the user knows we're working
	WinDrawChars("Retrieving Data...", 18, 10, 125);
	
	//* 2 - Package up the data and send it
	packet.buff = data;
	packet.len = Len; 
	dev_addr.u32 = 0xCCCBABE;
	if(IrTestReq( irLibRef, dev_addr, &con, &packet) == IR_STATUS_FAILED)
	{
		WinDrawChars("Send Failed!", 12, 10, 135);
	} 
}

void sendData2(char *data, int Len)
{
	//* 1 - Put a message on the screen so the user knows we're working
	WinDrawChars("Submiting Changes...", 20, 10, 125);
	
	//* 2 - Package up the data and send it
	packet.buff = data;
	packet.len = Len; 
	dev_addr.u32 = 0xCCCBABE;
	if(IrTestReq( irLibRef, dev_addr, &con, &packet) == IR_STATUS_FAILED)
	{
		WinDrawChars("Send Failed!", 12, 10, 130);
	} 

	sleep(5);
	WinDrawChars("                                        ", 40, 10, 125);
}

void openIR(void)
{
  IrStatus ret;

  //* open IR library
  if(SysLibFind(irLibName, &irLibRef)!=0) {
		FrmCustomAlert(kidAlert2, "", "", "");
  }  else {
    //* Open the library
    if(IrOpen(irLibRef, 0)!=0) {
			FrmCustomAlert(kidAlert2, "", "", "");
    } else {
			//* Now bind the callback funtion
 			if(IrBind(irLibRef, &con, IrCallback)!=IR_STATUS_SUCCESS) {
				FrmCustomAlert(kidAlert2, "", "", "");
 			} else {
				if(IrSetDeviceInfo(irLibRef, dev_name, 7)!=IR_STATUS_SUCCESS) {
					FrmCustomAlert(kidAlert2, "", "", "");
				} else {
	  			IrSetConTypeTTP(&con);  /* set connection type to TinyTP */

	    		ias_object.name = ias_name;
	    		ias_object.len  = StrLen(ias_name);
	    		ias_object.nAttribs = 1;
	    		ias_object.attribs = &attribs;

	    		attribs.name=attr_name;            /* Pointer to name of attribute */
       		attribs.len=StrLen(attr_name);     /* Length of attribute name */
       		attribs.value=attr_value;          /* Hardcode value (see below) */
       		attribs.valLen=5;                  /* Length of the value. */
	    		attr_value[4]=con.lLsap;

	    		// set up our service in IAS db
	    		if((ret=IrIAS_Add(irLibRef, &ias_object))!=IR_STATUS_SUCCESS)
					{
						FrmCustomAlert(kidAlert2, "", "", "");
					}
				}
			}
	  }
	}
}

void IrCallback(IrConnect *connect, IrCallBackParms *parms) {
  IrStatus ret;
  int i;
	char data[376];
  CALLBACK_PROLOGUE;

  switch(parms->event) {
  	case LEVENT_STATUS_IND:
			//* Get the connection parameters
    	break;
	
 		case LEVENT_TEST_IND:
			//* Handle the incomming data
			StrNCopy(data, parms->rxBuff, parms->rxLen);
			handleData(data, parms->rxLen);

			break;
			
  }
  CALLBACK_EPILOGUE;
}

void handleData(char *data, int Len)
{
	//* Get rid of the "Retrieving Data" sign
	WinDrawChars("                                        ", 40, 10, 125);

	//* Send data to the right guy
	switch (currentForm) {
		case Execute:
		case EExecute:
			//* Print the result of the command
			ExecuteHandle(data, Len);
			break;
		case ChangeIP:
		case PDEVICE0:
		case PDEVICE1:
			//* Decode the IP address information
			ChangeIPHandler(data, Len);
			break;
		case SMemory:
			//* Decode the Memory information
			SMemoryHandler(data, Len);
			break;
		case SLoadAvg:
			//* Decode the Load Average information
			SLoadHandler(data, Len);
			break;
		case STemperature:
			//* Decode the Temperature
			STemperatureHandler(data, Len);
			break;
	}
}

void ExecuteHandle(char *data, int len)
{
	/* Print the information out to the screen */
	data[len] = '\0';
	debug_msg(data);
}
void STemperatureHandler(char *data, int len)
{
	CharPtr find;
	
	//* Parse and print out the HI temperature
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 110, 50);

	//* Parse and print out the LO temperature
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 110, 60);

	//* Parse and print out the Current temperature
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 110, 70);

	//* Parse and print out the Fan setting
	data = find+1;
	WinDrawChars(data, 2, 110, 80);

}

void SMemoryHandler(char *data, int len)
{
	CharPtr find;
	
	//* Parse and print out the TOTAL MEMORY
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 50, 60);

	//* Parse and print out the USED MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 50, 70);

	//* Parse and print out the FREE MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 50, 80);

	//* Parse and print out the SHARED MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 50, 90);

	//* Parse and print out the BUFFERED MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 50, 100);

	//* Parse and print out the CACHED MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 50, 110);

	//* Parse and print out the TOTAL SWAP MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 100, 60);

	//* Parse and print out the USED SWAP MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 100, 70);

	//* Parse and print out the FREE SWAP MEMORY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	PrintNUMwCommas(data, StrLen(data), 100, 80);
}

void SLoadHandler(char *data, int len)
{
	CharPtr find;
	
	//* Parse and print out the SYSTEM TIME
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 100, 50);

	//* Parse and print out the UPTIME
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 100, 60);

	//* Parse and print out the NUMBER OF USERS
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 100, 70);

	//* Parse and print out the 1 MIN LOAD AVG
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 100, 80);

	//* Parse and print out the 5 MIN LOAD AVG
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 100, 90);

	//* Parse and print out the 15 MIN LOAD AVG
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	WinDrawChars(data, StrLen(data), 100, 100);
}

void PrintNUMwCommas(char *data, int Len, int x, int y)
{
	char output[32];
	int outputSofar = Len+4;
	int dataSofar = Len-1;
	int multiple = 0;

	//* Initialize the output as being spaces
	for(outputSofar = 0; outputSofar < Len + 4; outputSofar++)
		output[outputSofar] = ' ';

	//* Add in the commas
	while(dataSofar >= 0)
	{
		output[outputSofar--] = data[dataSofar--];

		if ((multiple == 2) && ((dataSofar+1) != 0)) { 
			output[outputSofar--]=',';
			multiple=0;
		} else {
			multiple++;
		}
	}
	outputSofar++;

	//* Print out the new number with the commas inserted
	WinDrawChars(output + outputSofar, (Len+4) - outputSofar + 1, x, y);
}

void ChangeIPHandler(char *data, int len)
{
	CharPtr find;
	FormPtr thisForm = FrmGetActiveForm();
	
	//* Parse and set the DEVICE
	find = StrChr(data, ',');
	find[0] = '\0';
	if (StrNCompare(data, "eth0", StrLen(data))==0)
	{
	  CtlSetValue(FrmGetObjectPtr(thisForm, FrmGetObjectIndex(thisForm, PDEVICE0)), 1);
	  CtlSetValue(FrmGetObjectPtr(thisForm, FrmGetObjectIndex(thisForm, PDEVICE1)), 0);
	} else if (StrNCompare(data, "eth1", StrLen(data))==0)
	{
	  CtlSetValue(FrmGetObjectPtr(thisForm, FrmGetObjectIndex(thisForm, PDEVICE0)), 0);
	  CtlSetValue(FrmGetObjectPtr(thisForm, FrmGetObjectIndex(thisForm, PDEVICE1)), 1);
	}

	//* Parse and print out the IP Address
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	SetFieldTextFromStr(FIPADDRESS, data);

	//* Parse and print out the NAME
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	SetFieldTextFromStr(FNAME, data);

	//* Parse and print out the ALIAS
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	SetFieldTextFromStr(FALIAS, data);

	//* Parse and print out the NETMASK
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	SetFieldTextFromStr(FNETMASK, data);

	//* Parse and print out the GATEWAY
	data = find+1;
	find = StrChr(data, ',');
	find[0] = '\0';
	SetFieldTextFromStr(FGATEWAY, data);
}

//* Allocates new handle and copies incoming string
//* This function was copied almost entirely from O'Reilly's Palm book
void SetFieldTextFromStr(Word fieldID, CharPtr strP)
{
  VoidHand txtH;
  VoidHand oldTxtH;
  FormPtr frm = FrmGetActiveForm();
  FieldPtr fldP;


  // get some space in which to stash the string.
  txtH = MemHandleNew(StrLen(strP) + 1);
  if (!txtH)
    return;

  // copy the string to the locked handle.
  StrCopy(MemHandleLock(txtH), strP);

  // unlock the string handle.
  MemHandleUnlock(txtH);

  // get the field and the field's current text handle.
  fldP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  ErrNonFatalDisplayIf(!fldP, "missing field");
  oldTxtH = (VoidHand) FldGetTextHandle(fldP);

  // set the field's text to the new text.
  FldSetTextHandle(fldP, (Handle)txtH);
  FldDrawField(fldP);

  // free the handle AFTER we call FldSetTextHandle().
  if (oldTxtH)
    MemHandleFree(oldTxtH);
}

void init_buffer(void) {
  int i;

  text_height=FntLineHeight()*TEXT_LINES;

	disp_line = 0;

  /* clear buffer */
  for(i=0;i<MAX_LINES;i++)
    text_buffer[i][0]=0;
}

void debug_msg(char *msg) {
  RectangleType rect; 
  int i, j, len;

	len = StrLen(msg);
	for(j = 0; j < len; j++)
	{
		switch(msg[j]) {
			case 8 : // Deal with a backspace
				if(text_buffer[0][0] > 0)
				{
  				text_buffer[0][0]--;

					//* Redraw the current line
    			//* clear text area 
    			rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP+text_height-FntLineHeight(); 
    			rect.extent.x=160; rect.extent.y=FntLineHeight();
    			WinEraseRectangle(&rect,0);
					//* Redraw the line
    			WinDrawChars(&text_buffer[0][1], text_buffer[0][0], 
						0, TEXT_TOP+text_height-FntLineHeight());
				}
				break;
			case '\n' : //* Deal with a new line

				//* Redraw the current line
    		//* clear text area 
    		rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP+text_height-FntLineHeight(); 
    		rect.extent.x=160; rect.extent.y=FntLineHeight();
    		WinEraseRectangle(&rect,0);
				//* Redraw the line
    		WinDrawChars(&text_buffer[disp_line][1], text_buffer[disp_line][0], 
						0, TEXT_TOP+text_height-FntLineHeight());
				
  			/* scroll buffer up one line */
  			for(i=MAX_LINES;i>0;i--)
    			MemMove(text_buffer[i], text_buffer[i-1], MAX_LEN+1);

    		/* scroll text up */
    		rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP+FntLineHeight(); 
    		rect.extent.x=160; rect.extent.y=text_height-FntLineHeight();
    		WinCopyRectangle(draw_win, draw_win, &rect, 0, TEXT_TOP, scrCopy);
  
    		/* clear new text area */
    		rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP+text_height-FntLineHeight(); 
    		rect.extent.x=160; rect.extent.y=FntLineHeight();
    		WinEraseRectangle(&rect,0);

  			text_buffer[0][0] = 0;
	
				break;
			default:
				//* Store the Character
				if(text_buffer[0][0] < MAX_LEN)
				{
  				text_buffer[0][0]++;
  				text_buffer[0][text_buffer[0][0]] = msg[j];
				}
		}
	}
	
  if(draw_win==WinGetActiveWindow()) {
    /* draw string */
		text_buffer[0][text_buffer[0][0]+1] = 0;
    WinDrawChars(&text_buffer[disp_line][1], text_buffer[disp_line][0]+1, 0, TEXT_TOP+text_height-FntLineHeight());
  }
}

void EAScrollUp(void) 
{
  RectangleType rect; 
  int i;

	if(disp_line + TEXT_LINES < MAX_LINES)
	{
		disp_line++;

  	if(draw_win==WinGetActiveWindow()) {
    	/* scroll text up */
    	rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP; 
    	rect.extent.x=160; rect.extent.y=text_height-FntLineHeight();
    	WinCopyRectangle(WinGetActiveWindow(),WinGetActiveWindow(), &rect, 0, TEXT_TOP+FntLineHeight(), scrCopy);
    	
    	/* clear new text area */
    	rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP; 
    	rect.extent.x=160; rect.extent.y=FntLineHeight();
    	WinEraseRectangle(&rect,0);
    	
    	
    	/* draw string */
    	WinDrawChars(&text_buffer[disp_line+TEXT_LINES-1][1], text_buffer[disp_line+TEXT_LINES-1][0], 0, TEXT_TOP);
  	}
	} 
}

void EAScrollDown(void) 
{
  RectangleType rect; 
  int i;

	if(disp_line > 0)
	{
		disp_line--;

  	if(draw_win==WinGetActiveWindow()) {
    	/* scroll text down */
    	rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP+FntLineHeight(); 
    	rect.extent.x=160; rect.extent.y=text_height-FntLineHeight();
    	WinCopyRectangle(WinGetActiveWindow(),WinGetActiveWindow(), &rect, 0, TEXT_TOP, scrCopy);
    	
    	/* clear new text area */
    	rect.topLeft.x=0;  rect.topLeft.y=TEXT_TOP+text_height-FntLineHeight(); 
    	rect.extent.x=160; rect.extent.y=FntLineHeight();
    	WinEraseRectangle(&rect,0);
    	
    	/* draw string */
    	WinDrawChars(&text_buffer[disp_line][1], text_buffer[disp_line][0], 0, TEXT_TOP+text_height-FntLineHeight());
  	} 
	}
}

